require("serialize")
Saver = require("saves")
require("gui-menu")
require("console")

Loader = {}
Loader.bind = {}
GUI = require("gui").create()

dbg = require("debug_tools")

Loader.slaveKeyProcessors = {}

function Loader.start()
	GlobalSetKeyDownProc( Loader.keyProcessor )
    Menu.showMainMenu()
end

function Loader.gameInit()
    Loader.old_time = GetCurTime()
    Loader.dt = 0
    Loader.time = 0
    Log("Init time ", Loader.time, "  old time ", Loader.old_time)
    GUI:show()

    SetOnPlayerDeathProcessor(Loader.onPlayerDeath)    -- ,      
    if CharacterSystem then
        SetOnChangePlayerProcessor(CharacterSystem);
    end
end

function Loader.startGame(map_name)
    if Loader.level and Loader.level.cleanUp then Loader.level.cleanUp() end
	
	Loader.map_name = map_name
    godmode = {}
    stat = { lives = 3, score = 0 }
    mapvar = { lives = 3, score = 0, tmp = {} }
    local map_script = "scripts/"..map_name..".lua"
    Log('loading game:', map_script)
	local result, success = loadfile(map_script)
	if result then
		success, result = pcall(result)
	else	
		result = success
		success = nil
	end
	if success then
		Loader.level = result
		Loader.level.SetLoader(Loader)
		-- starting mission
		Loader.state = 'new'

		Loader.gameInit()
    
		if not Loader.missionThread then 
			Loader.missionThread = NewThread( Loader.mainLoop );
			Resume(Loader.missionThread);
		end
		return true, "Map loaded."
	else
		return false, "Failed to load map: "..result
	end
end

function Loader.restartGame()    
    --  hiding irrelevant widgets
    StopMapThreads()
    if Loader.level and Loader.level.cleanUp then Loader.level.cleanUp() end
	if GUI.game_over_widgets then GUI.game_over_widgets:show(false) end
    Menu:hide()

    mapvar = shallow_copy(stat)

    TogglePause();
    Menu.mn = false;
    mapvar.lives = stat.lives or 3;
        
    Loader.level = dofile("scripts/"..Loader.map_name..".lua")
    Loader.level.SetLoader(Loader)
    
    Loader.gameInit()
end

--       new_level
function Loader.ChangeLevel(new_level)
    mapvar.threads = nil    --   .
    mapvar.secrets = nil    --    -  .
    mapvar.tmp = {}         --     .
    stat = shallow_copy( mapvar )
    
    local fplr = GetPlayer(1)
    local splr = GetPlayer(2)
    
    Loader.map_name = new_level

    Loader.restartGame()
    
    if fplr then
        SetPlayerAltWeapon(1, fplr.altWeaponName)
        SetPlayerAmmo(1, fplr.ammo)
        SetPlayerHealth(1, fplr.health)
        if ( fplr.altWeaponActive ) then SwitchWeapon( 1 ) end
    end
    
    if splr then
        SetPlayerAltWeapon(2, splr.altWeaponName)
        SetPlayerAmmo(2, splr.ammo)
        SetPlayerHealth(2, splr.health)
        if splr.active then
            SwitchCharacter(true)
        end
        if ( splr.altWeaponActive ) then SwitchWeapon( 2 ) end
    end
end

function Loader.exitGame()
    -- destroying all widgets
    StopMapThreads()
    StopBackMusic()
    GUI:hide()
    GUI.gc:destroy()
    GUI.widgetContainer:destroy()
    GUI.game_over_widgets:destroy()
    Menu.mn = false
    GUI:destroy()
    if not mapvar or  not mapvar.score then
        mapvar.score = 0;
        mapvar.lives = 3;
    end
    -- leaving mission
	if Loader.level and Loader.level.cleanUp then Loader.level.cleanUp() end
	Loader.level = nil
	Loader.state = ''    --idling mainLoop thread
    DestroyGame()        -- destroying level
    Menu.showMainMenu()
end

------------------------------- main script game loop  ---------------------------------------------------
function Loader.mainLoop()
    local kp
    while true do
	kp = GlobalGetKeyReleaseProc()
        if Loader.state == 'new' then
            Loader.state = 'game'
            Loader.gameOverShown = false
		Menu.goshown = false
            GUI:init()
        elseif Loader.state == 'game' then
            Loader.ingame()
        elseif Loader.state == 'game over' then
            Loader.showGameOver()
        end
        Wait(1)
    end
end

function Loader.keyProcessor( key )
	if key == keys["tilde"] and not Console.shown then
		Console.toggle()
	elseif Loader.state == 'game' and key == CONFIG.key_conf[1].gui_nav_menu and ( not Menu.locked or not Menu.locked[#Menu.locked] ) then
		Menu.mn = not Menu.mn
	        if Menu.mn then
		    push_pause( true )
	            Menu:show(3)
        	else
		    pop_pause()
        	    Menu:hide()
	        end		
	end
	for tkey, value in pairs(Loader.slaveKeyProcessors) do
		value( key )
	end
	if Loader.bind[key] then Loader.bind[key]() end
end

function Loader.ingame()
    local player = GetPlayer()
    local hp = iff( player and player.health, player.health, 0 )
    if not Editor or not Editor.active then
	    Loader.level.MapScript(player)
    end
    Loader.omb = ButtonPressed(CONFIG.key_conf[1].gui_nav_menu)
    if Heartbeat and not Editor or not Editor.active then Heartbeat() end

	if dbg then dbg.process() end
	
    local t = GetCurTime()
    if not ( GamePaused() ) then
        Loader.dt = t - Loader.old_time
         Loader.time = Loader.time + Loader.dt
    end
     Loader.old_time = t
end

-- ,   ,     ,   GetPlayer   
-- old_plr_info -      . num -   (1  2)
function Loader.onPlayerDeath(old_plr_info, num)
    mapvar.lives = mapvar.lives - 1
    mapvar.sync = 0
    if mapvar.lives > 0 then
        local plr_id = RevivePlayer(num, nil);
        SetCamAttachedObj(plr_id)
        --     
        SetPlayerAltWeapon(num, old_plr_info.altWeaponName)
        SetPlayerAmmo(num, old_plr_info.ammo)
        inv = num
        if ( old_plr_info.altWeaponActive ) then
            SwitchWeapon( num )
        end
    else
        Loader.state = 'game over'
    end
end

function Loader.showGameOver()
	if not Menu.goshown then
		Menu.goshown = true
		Menu.showGameOver()
		Menu:show(7)
	end
end


return Loader
