pause_stack = {}

function push_pause( state )
	table.insert( pause_stack, GamePaused() )
	TogglePause( state )
end

function pop_pause()
	TogglePause( (pause_stack[#pause_stack] or false) )
	table.remove( pause_stack, #pause_stack )
end

function dictionary_string(string)
	if not dictionary or not dictionary[string] then
		return string
	end
	return dictionary[string]
end

function iff( condition, a, b )
	if condition then
		return a
	else
		return b
	end
end

function diff_iff( condition, object )
	if condition then
		return object
	else
		return { constants.ObjNone }
	end
end

function child( parent )
	local a = {}
	local a_mt = { __index = a }
	
	function a.bareCreate()
		local ni = {}
		setmetatable( ni, a_mt )
		return ni
	end
	
	function a.superCreate( ... )
		local b = parent.create( unpack(arg) )
		setmetatable( b, a_mt ) 
		return b
	end
	
	function a:super( func, ... )
		local l_arg = arg
		table.remove(l_arg, 1)
		return parent[func]( parent, l_arg )
	end
	
	if parent then
		setmetatable( a, { __index = parent } )
		a.super = parent
	end
	
	return a
end

function shallow_copy( source )
	if not source then return {} end
	local ret = {}
	for key, value in pairs(source) do
		ret[key] = value
	end
	return ret
end

function is_char( char )
	if not char then return false end
	if #char == 1 then
		return true
	end
	return false
end

function random( tab )
	if tab then
		if #tab == 0 then
			ret = {}
			for key, value in pairs(tab) do
				table.insert(ret, value)
			end
		else
			ret = tab
		end
		return ret[math.random(1, #ret)]
	end
	return nil
end

function protected_string(str)
	return string.gsub(str, "[%.%!%*%?%-%[%]%(%)%^%$%%]", "%%%1")
end

local function nextid()
	WidgetContainer.last_id = WidgetContainer.last_id + 1
	return WidgetContainer.last_id - 1
end

function NewMapThread( thread )
	if not mapvar then
		return NewThread( thread, true )
	else
		if not mapvar.threads then mapvar.threads = {} end
		local ret = NewThread( thread, true )
		table.insert( mapvar.threads, ret )
		return ret
	end
end

function StopMapThreads()
	if not mapvar or not mapvar.threads then return end
	for key, value in pairs( mapvar.threads ) do
		StopThread( value )
	end
	mapvar.threads = {}
end

----------------------------------------------------------------------------
----------------------------------------------------------------------------
-- WidgetContainer

WidgetContainer = {}
WidgetContainer.__index = WidgetContainer
WidgetContainer.w2c = {}
WidgetContainer.w2id = {}
WidgetContainer.last_id = 1

function WidgetContainer.create(id, visibility, on_show)
	local a = {}
	setmetatable(a,WidgetContainer)
	a.elements = {}
	a.pointers = {}
	a.onfocus = {}
	a.onunfocus = {}
	a.params = {}
	a.onleft = {}
	a.onright = {}
	a.onkey = {}
	a.visible = {}
	a.id = id
	a.visibility = visibility or false
	a.on_show = on_show
	return a
end

function WidgetContainer:addLabel(WidgetTable)
	local wt = WidgetTable
	wt.text = wt.text or "NULL"
	if dictionary and dictionary[wt.text] then wt.text = dictionary[wt.text] end
	wt.color = wt.color or {1, 1, 1, 1}
	wt.x = wt.x or 0
	wt.y = wt.y or 0
	wt.id = wt.id or "WIDGET" ..nextid()
	wt.font = wt.font or "default"
	wt.visible = wt.visible or true
	if wt.center and wt.text then
		local dx, dy = GetCaptionSize(wt.font, wt.text)
		--Log(wt.text.." "..dx.." "..dy)
		wt.x = wt.center[1]-math.floor(dx/2)
		wt.y = wt.center[2]-math.floor(dy/2)
		--Log(wt.text.." "..wt.x.." "..wt.y)
	end
	local widget = CreateWidget(constants.wt_Label, wt.id, nil, wt.x, wt.y, GetCaptionSize(wt.font, wt.text))
	WidgetSetCaption(widget, wt.text)
	WidgetSetCaptionColor(widget, wt.color, false)
	WidgetSetCaptionFont(widget, wt.font)
	WidgetSetVisible(widget, self.visibility and wt.visible)

	if self.elements[wt.id] then DestroyWidget(self.elements[wt.id]) end
	self.elements[wt.id] = widget
	self.visible[wt.id] = wt.visible
	WidgetContainer.w2c[wt.id] = self
	WidgetContainer.w2id[widget] = wt.id

	return wt.id
end

function WidgetContainer:addPicture(WidgetTable)
	local wt = WidgetTable
	wt.sprite = wt.sprite or "wakaba-widget"
	wt.anim = wt.anim or "idle"
	wt.visible = iff(wt.invisible, false, true)
	wt.x = wt.x or 0
	wt.y = wt.y or 0
	wt.w = wt.w or 28
	wt.h = wt.h or 28
	wt.id = wt.id or "WIDGET" ..nextid()
	local widget = CreateWidget(constants.wt_Picture, wt.id, nil, wt.x, wt.y, wt.w, wt.h)
	WidgetSetSprite(widget, wt.sprite)
	WidgetSetAnim(widget, wt.anim)
	WidgetSetLMouseClickProc(widget, WidgetContainer.onLeft)
	WidgetSetRMouseClickProc(widget, WidgetContainer.onRight)
	WidgetSetFocusProc(widget, WidgetContainer.onFocus)
	WidgetSetUnFocusProc(widget, WidgetContainer.onUnfocus)
	WidgetSetKeyPressProc(widget, WidgetContainer.onKey)
	WidgetSetVisible(widget, (self.visibility and wt.visible))

	if self.elements[wt.id] then DestroyWidget(self.elements[wt.id]) end
	self.elements[wt.id] = widget
	self.onfocus[wt.id] = wt.on_focus
	self.onunfocus[wt.id] = wt.on_unfocus
	self.onleft[wt.id] = wt.on_left
	self.onright[wt.id] = wt.on_right
	self.onkey[wt.id] = wt.on_key
	self.params[wt.id] = wt.params
	self.visible[wt.id] = wt.visible
	WidgetContainer.w2c[wt.id] = self
	WidgetContainer.w2id[widget] = wt.id

	return wt.id
end

function WidgetContainer:addButton(WidgetTable)
	local wt = WidgetTable
	wt.font = wt.font or "default"
	wt.color = wt.color or {1, 1, 1, 1}
	wt.color_focused = wt.color_focused or {1, 1, 1, 1}
	wt.anim = wt.anim or "idle"
	wt.visible = wt.visible or true
	wt.x = wt.x or 0
	wt.y = wt.y or 0
	if wt.text and dictionary and dictionary[wt.text] then wt.text = dictionary[wt.text] end
	local sz1, sz2 = GetCaptionSize(wt.font, (wt.text) or ".")
	wt.w = wt.w or sz1
	wt.h = wt.h or sz2
	wt.id = wt.id or "WIDGET" ..nextid()
	local dx, dy = 0,0
	if wt.center and wt.text then
		dx = wt.x - (wt.center[1]-math.floor(sz1/2))
		dy = wt.y - (wt.center[2]-math.floor(sz2/2))
		wt.x = wt.x - dx
		wt.y = wt.y - dy
	end
	local widget = CreateWidget(constants.wt_Button, wt.id, nil, wt.x, wt.y, wt.w, wt.h)
	if wt.text then
		--Log(wt.text..":"..(dictionary[wt.text] or "nil"))
		WidgetSetCaption(widget, wt.text)
		WidgetSetCaptionColor(widget, wt.color, false)
		WidgetSetCaptionColor(widget, wt.color_focused, true)
		WidgetSetCaptionFont(widget, wt.font)
	end
	WidgetSetVisible(widget, (self.visibility and wt.visible))
	if wt.sprite then
		WidgetSetSprite(widget, wt.sprite)
		WidgetSetAnim(widget, wt.anim)
	end
	WidgetSetLMouseClickProc(widget, WidgetContainer.onLeft)
	WidgetSetRMouseClickProc(widget, WidgetContainer.onRight)
	WidgetSetFocusProc(widget, WidgetContainer.onFocus)
	WidgetSetUnFocusProc(widget, WidgetContainer.onUnfocus)
	if wt.non_focusable then
		WidgetSetFocusable(widget, false)
	end

	if wt.pointer_proto then
		pointer = CreateWidget(constants.wt_Picture, wt.id.."_pointer", nil, wt.pointer_x-dx, wt.pointer_y-dy, 1, 1)
		WidgetSetSprite(pointer, wt.pointer_proto)
		WidgetSetAnim(pointer, "idle")
		WidgetSetVisible(pointer, false)
	end

	if self.elements[wt.id] then DestroyWidget(self.elements[wt.id]) end
	if self.pointers[wt.id] then 
		--Log(wt.id)
		--Log(self.pointers[wt.id])
		DestroyWidget(self.pointers[wt.id]) 
		self.pointes[wt.id] = nil 
	end
	self.elements[wt.id] = widget
	self.onfocus[wt.id] = wt.on_focus
	self.onunfocus[wt.id] = wt.on_unfocus
	self.onleft[wt.id] = wt.on_left
	self.onright[wt.id] = wt.on_right
	self.params[wt.id] = wt.params
	self.pointers[wt.id] = pointer
	self.visible[wt.id] = wt.visible
	WidgetContainer.w2c[wt.id] = self
	WidgetContainer.w2id[widget] = wt.id

	--Log(serialize("wt", wt))

	return wt.id
end

function WidgetContainer:show(state)
	Log(state or "nil")
	if self.on_show then
		self.on_show(state)
	end
	self.visibility = state
	for key, value in pairs(self.elements) do
		WidgetSetVisible(self.elements[key], (self.visible[key] and self.visibility))
		if not self.visibility and self.pointers[key] then WidgetSetVisible(self.pointers[key], false) end
	end
end

function WidgetContainer:destroy()
	for key, value in pairs(self.elements) do
		DestroyWidget(self.elements[key])
		if self.pointers[key] then 
			DestroyWidget(self.pointers[key]) 
		end
		self.elements[key] = nil
		self.pointers[key] = nil
	end
end

function WidgetContainer:destroyWidget(id)
	if self.elements[id] then
		DestroyWidget(self.widgets[id])
		self.elements[id] = nil
		if self.pointers[id] then
			DestroyWidget(self.pointers[id])
		end
	else
		Log(string.format("WARNING: WidgetContainer %s: trying to destroy widget %s which does not exists.", (self.id or "BAD CONTAINER"), (id or "BAD WIDGET")))
	end
end

function WidgetContainer:showWidget(id, state)
	if self.elements[id] then
		self.visible[id] = iff(state, true, false)
		WidgetSetVisible(self.elements[id], (self.visible[id] and self.visibility))
		if not self.visible[id] and self.pointers[id] then WidgetSetVisible(self.pointers[id], false) end
	else
		Log(string.format("WARNING: WidgetContainer %s: trying to change visibility of widget %s which does not exists.", (self.id or "BAD CONTAINER"), (id or "BAD WIDGET")))
	end

end

function WidgetContainer.onFocus(sender)
	local id = WidgetContainer.w2id[sender]
	local cont = WidgetContainer.w2c[id]
	if not cont or not id then
		Log("ERROR: A call to WidgetContainer function from widget not in the container!")
		return
	end
	
	if ( cont.pointers[id] ) then
		WidgetSetVisible( cont.pointers[id], true )
	end
	if ( cont.onfocus[id] ) then
		cont.onfocus[id](sender, cont.params[id])
	end
end

function WidgetContainer.onUnfocus(sender)
	local id = WidgetContainer.w2id[sender]
	local cont = WidgetContainer.w2c[id]

	if not cont or not id then
		Log("ERROR: A call to WidgetContainer function from widget not in the container!")
		return
	end
	
	if ( cont.pointers[id] ) then
		WidgetSetVisible( cont.pointers[id], false )
	end
	if ( cont.onunfocus[id] ) then
		cont.onunfocus[id](sender, cont.params[id])
	end
end

function WidgetContainer.onLeft(sender)
	local id = WidgetContainer.w2id[sender]
	local cont = WidgetContainer.w2c[id]

	if not cont or not id then
		Log("ERROR: A call to WidgetContainer function from widget not in the container!")
		return
	end
	
	if ( cont.onleft[id] ) then
		cont.onleft[id](sender, cont.params[id])
	end
end

function WidgetContainer.onRight(sender)
	local id = WidgetContainer.w2id[sender]
	local cont = WidgetContainer.w2c[id]

	if not cont or not id then
		Log("ERROR: A call to WidgetContainer function from widget not in the container!")
		return
	end
	
	if ( cont.onright[id] ) then
		cont.onright[id](sender, cont.params[id])
	end
end

function WidgetContainer.onKey(sender, key)
	local id = WidgetContainer.w2id[sender]
	local cont = WidgetContainer.w2c[id]

	if not cont or not id then
		Log("ERROR: A call to WidgetContainer function from widget not in the container!")
		return
	end
	
	if ( cont.onkey[id] ) then
		cont.onkey[id](sender, key)
	end
end

----------------------------------------------------------------------------
----------------------------------------------------------------------------
-- Window

Window = {}
Window.__index = Window
Window.widgets = {}

local function CWW(x, y, anim, tab, sprite)
	local widget = CreateWidget(constants.wt_Picture, "window", nil, x, y, -32, -32)
	WidgetSetSprite(widget, sprite)
	WidgetSetVisible(widget, true)
	WidgetSetAnim(widget, anim)
	WidgetSetFocusable(widget, false)
	table.insert(tab, widget)
end

function Window:prepareWidgets()
	local tab = {}
	tab.top = math.max(math.floor( (self.w-32)/16 + 1 ), 1 )
	tab.bottom = tab.top
	tab.left = math.max(math.floor( (self.h-32)/16 + 1 ), 1 )
	tab.right = tab.left
	tab.center = tab.top * tab.left
	for key, value in pairs(tab) do
		for i=#self.widgets[key]+1,value do
			--Log( "i=", i, ", key=", key )
			local widget = CreateWidget( constants.wt_Picture, "window", nil, -32, -32, 16, 16 )
			self.widgets[key][i] = widget
			WidgetSetSprite(widget, self.sprite)
			WidgetSetVisible(widget, true)
			WidgetSetAnim(widget, key)
			WidgetSetFocusable(widget, false)
		end
		for i=value+1,#self.widgets[key] do
			WidgetSetVisible(self.widgets[key][i], false)
		end
		for i=1,value do
			WidgetSetVisible(self.widgets[key][i], true)
		end
	end
	if #self.widgets.corners == 0 then
		CWW(0, 0, "topleft", self.widgets.corners, self.sprite)
		CWW(0, 0, "topright", self.widgets.corners, self.sprite)
		CWW(0, 0, "bottomright", self.widgets.corners, self.sprite)
		CWW(0, 0, "bottomleft", self.widgets.corners, self.sprite)
	end
end

function Window:show()
	--self:setVisible(true)
	local tab = { side = 1, updown = 1, center = 1 }
	for i=self.x+16,self.x+self.w-32,16 do
		WidgetSetPos( self.widgets.top[tab.updown], i, self.y )
		for j=self.y+16,self.y+self.h-32,16 do
			WidgetSetPos( self.widgets.center[tab.center], i, j )
			tab.center = tab.center + 1
		end
		WidgetSetPos( self.widgets.center[tab.center], i, self.y+self.h-32 )
		tab.center = tab.center + 1
		WidgetSetPos( self.widgets.bottom[tab.updown], i, self.y+self.h-16 )
		tab.updown = tab.updown + 1
	end
	
	WidgetSetPos( self.widgets.top[tab.updown], self.x+self.w-32, self.y )
	for j=self.y+16,self.y+self.h-32,16 do
		WidgetSetPos( self.widgets.center[tab.center], self.x+self.w-32, j )
		tab.center = tab.center + 1
	end
	WidgetSetPos( self.widgets.center[tab.center], self.x+self.w-32, self.y+self.h-32 )
	tab.center = tab.center + 1
	WidgetSetPos( self.widgets.bottom[tab.updown], self.x+self.w-32, self.y+self.h-16 )
	tab.updown = tab.updown + 1
	
	for j=self.y+16,self.y+self.h-32,16 do
		WidgetSetPos( self.widgets.left[tab.side], self.x, j )
		WidgetSetPos( self.widgets.right[tab.side], self.x+self.w-16, j )
		tab.side = tab.side + 1
	end
	WidgetSetPos( self.widgets.left[tab.side], self.x, self.y+self.h-32 )
	WidgetSetPos( self.widgets.right[tab.side], self.x+self.w-16, self.y+self.h-32 )
	
	for i = 1,4 do
		WidgetSetVisible( self.widgets.corners[i], true )
	end
	WidgetSetPos( self.widgets.corners[1], self.x, self.y )
	WidgetSetPos( self.widgets.corners[2], self.x+self.w-16, self.y )
	WidgetSetPos( self.widgets.corners[3], self.x+self.w-16, self.y+self.h-16 )
	WidgetSetPos( self.widgets.corners[4], self.x, self.y+self.h-16 )
	for i=tab.side+1,#self.widgets.left do
		WidgetSetVisible(self.widgets.left[i], false)
		WidgetSetVisible(self.widgets.right[i], false)
	end
	for i=tab.updown+1,#self.widgets.top do
		WidgetSetVisible(self.widgets.top[i], false)
		WidgetSetVisible(self.widgets.bottom[i], false)
	end
	for i=tab.center+1,#self.widgets.center do
		WidgetSetVisible(self.widgets.center[i], false)
	end
end

function Window:resize( w, h )
	self.w = w
	self.h = h
	self:prepareWidgets()
	self:show()
end

function Window:move( x, y )
	self.x = x
	self.y = y
	self:show()
end

function Window:recreate( x, y, w, h )
	self.x = x
	self.y = y
	self.w = w
	self.h = h
	self:prepareWidgets()
	self:show()
end

function Window.create(x, y, w, h, sprite)
	local a = {}
	setmetatable(a, Window)
	a.x = x
	a.y = y
	a.w = w
	a.h = h
	a.sprite = sprite
	a.visible = true
	a.widgets = { top = {}, bottom = {}, left = {}, right = {}, center = {}, corners = {} }
	a:prepareWidgets()
	a:show()
	return a
end

function Window:setVisible(visibility)
	self.visible = visibilty
	if not self.visible then
		for key, value in pairs(self.widgets) do
			for key2, value2 in pairs(value) do
				WidgetSetVisible(value2, false)
			end
		end
	else
		self:prepareWidgets()
	end
end

function Window:destroy()
	for key, value in pairs(self.widgets) do
		for key2, value2 in pairs(value) do
			DestroyWidget(value2)
		end
	end
end

function LetterBox( state )
	if state then
		letterbox1 = CreateWidget(constants.wt_Widget, "letterbox1", nil, 0, 0, CONFIG.scr_width, CONFIG.scr_height/8)
		WidgetSetZ(letterbox1, 0.5)
		WidgetSetColorBox( letterbox1, {0,0,0,1} )
		letterbox2 = CreateWidget(constants.wt_Widget, "letterbox2", nil, 0, CONFIG.scr_height*7/8, CONFIG.scr_width, CONFIG.scr_height/8)
		WidgetSetZ(letterbox2, 0.5)
		WidgetSetColorBox( letterbox2, {0,0,0,2} )
	else
		if ( letterbox1 ) then DestroyWidget(letterbox1) end
		if ( letterbox2 ) then DestroyWidget(letterbox2) end
	end
end

----------------------------------------------------------------------------
----------------------------------------------------------------------------
-- Fading

local fade_color = {0, 0, 0, 0}
local fade_box = nil
local fade_z = 0.5
local fading_in = false

local function FadeOutThread( color, time )
	--Log("FadeOutThread called")
	if not color then color = { 0, 0, 0 } end
	if time == 0 then time = 1 end
	local fade_start = GetCurTime()
	local fade_alpha = 0
	local progress
	fade_color = color
	
	if not fade_box then
		fade_box = CreateWidget(constants.wt_Widget, "fade_box", nil, 0, 0, CONFIG.scr_width, CONFIG.scr_height)
		WidgetSetZ(fade_box, fade_z)
		WidgetSetVisible(fade_box, true)
	end	
	
	while not fading_in do
		progress = GetCurTime() - fade_start
		if progress > time then fade_alpha = 0.2 else
			fade_alpha = (progress/time)*0.2
		end
		WidgetSetColorBox(fade_box, {color[1], color[2], color[3], fade_alpha} )
		Wait(1)
	end
	
	--Log("FadeOutThread ending, fading_in = ", fading_in)
end

local function FadeInThread( time )
	local fade_start = GetCurTime()
	--Log("FadeInThread called, time = ", time, ", cur_time = ", fade_start, ", fading_in = ", fading_in, ", fade_box = ", fade_box)
	if not fade_box then return end
	local fade_alpha = 1
	while (fade_start + time > GetCurTime()) and fading_in do
		fade_alpha = 1 - (GetCurTime()-fade_start)/time
		WidgetSetColorBox(fade_box, {fade_color[1], fade_color[2], fade_color[3], fade_alpha} )
		Wait(1)
	end
	if fading_in then 
		DestroyWidget(fade_box)
		fade_box = nil
	end	
	--Log("FadeInThread ending, fading_in = ", fading_in)
	fading_in = false
end

function FadeOut( color, time )
	--Log("FadeOut called, fading_in = ", fading_in)
	fading_in = false
	if time > 0 then		
		local fade_thread = NewThread( FadeOutThread )
		Resume( fade_thread, color, time )
	else
		screen = GetScreenInfo()
		if not fade_box then 
			fade_box = CreateWidget(constants.wt_Widget, "fade_box", nil, 0, 0, CONFIG.scr_width, CONFIG.scr_height)
		end
		WidgetSetZ(fade_box, fade_z)
		WidgetSetVisible(fade_box, true)
		WidgetSetColorBox(fade_box, {color[1], color[2], color[3], 1} )
	end
end

function FadeIn( time )
	--Log("FadeIn called, fading_in = ", fading_in)
	if fading_in then return end
	fading_in = true
	local fade_thread = NewThread( FadeInThread )
	Resume( fade_thread, time )
end

----------------------------------------------------------------------------
----------------------------------------------------------------------------
local function MoveWidgetToThread(widget, new_pos, time, linear, onEnd)
	local old_x, old_y = WidgetGetPos(widget)
	if not old_x or not old_y then return end
	local time_start = GetCurTime()
	local cur_time = time_start
	
	local x = 0
	local y = 0
	
	while time_start + time > cur_time do
		local t = (cur_time - time_start)/time

		if not linear then t = t*t*(3-2*t) end
		
		x = old_x + (new_pos[1] - old_x) * t
		y = old_y +(new_pos[2] - old_y) * t
				
		WidgetSetPos(widget, x, y)
		
		Wait(1)
		cur_time = GetCurTime()
	end 
	
	WidgetSetPos(widget, new_pos[1], new_pos[2])	
	if onEnd and (type(onEnd) == "function") then onEnd() end
	Log("MoveWidgetToThread ended")
end

function MoveWidgetTo(widget, new_pos, time, linear, onEnd)
	if (not new_pos) or (type(new_pos) ~= "table") then return end
	local move_thread = NewThread(MoveWidgetToThread)
	Log("move_thread = ", move_thread)
	Resume(move_thread, widget, new_pos, time, linear, onEnd)
end

----------------------------------------------------------------------

function isConfigKeyPressed(key, cfgKeyName)
	for i = 1, constants.configKeySetsNumber do
		if CONFIG.key_conf[i][cfgKeyName] == key then return true end
	end
	return false
end
--------------------------------------------------------------------------

routines = {}

return routines
