#include "StdAfx.h"

#include "../config.h"

#include "camera.h"
#include "objects/object_player.h"

//////////////////////////////////////////////////////////////////////////

extern config cfg;

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

//  
float CAMERA_X = 0;
float CAMERA_Y = 0;

//   
float CAMERA_LEFT = 0;
float CAMERA_RIGHT = 0;
float CAMERA_TOP = 0;
float CAMERA_BOTTOM = 0;

//    
float CAMERA_OFF_X = 0;
float CAMERA_OFF_Y = 0;

float CAMERA_OLD_OFF_X = 0;
float CAMERA_OLD_OFF_Y = 0;

//    
float CAMERA_OBJ_OFF_X = 0;
float CAMERA_OBJ_OFF_Y = 0;

//  ,    
CameraFocusObjectPoint CAMERA_FOCUS_ON_OBJ_POS = CamFocusCenter;

// ,    
GameObject* attached_object = NULL;

bool camera_attach_x = true;
bool camera_attach_y = true;

bool camera_use_bounds = false;
float CAMERA_LEFT_BOUND = 0;
float CAMERA_RIGHT_BOUND = 0;
float CAMERA_TOP_BOUND = 0;
float CAMERA_BOTTOM_BOUND = 0;
//////////////////////////////////////////////////////////////////////////
// 

__INLINE void CalcCamOffX()
{
	CAMERA_OFF_X = -(CAMERA_X - cfg.scr_width * 0.5f);
	CAMERA_LEFT = -CAMERA_OFF_X;
	CAMERA_RIGHT = CAMERA_LEFT + cfg.scr_width;
}

__INLINE void CalcCamOffY()
{
	CAMERA_OFF_Y = -(CAMERA_Y - cfg.scr_height * 0.5f);
	CAMERA_TOP = -CAMERA_OFF_Y;
	CAMERA_BOTTOM = CAMERA_TOP + cfg.scr_height;
}

__INLINE void ApplyBounds()
{
	if (CAMERA_LEFT < CAMERA_LEFT_BOUND)
	{
		CAMERA_X += CAMERA_LEFT_BOUND - CAMERA_LEFT;
		CalcCamOffX();
	}

	if (CAMERA_RIGHT >  CAMERA_RIGHT_BOUND)
	{
		CAMERA_X -= CAMERA_RIGHT - CAMERA_RIGHT_BOUND;
		CalcCamOffX();
	}

	if (CAMERA_TOP < CAMERA_TOP_BOUND)
	{
		CAMERA_Y += CAMERA_TOP_BOUND - CAMERA_TOP;
		CalcCamOffY();
	}

	if (CAMERA_BOTTOM >  CAMERA_BOTTOM_BOUND)
	{
		CAMERA_Y -= CAMERA_BOTTOM - CAMERA_BOTTOM_BOUND;
		CalcCamOffY();
	}
}
//////////////////////////////////////////////////////////////////////////
void CameraAttachToAxis(bool x, bool y)
{
	camera_attach_x = x;
	camera_attach_y = y;
}

//    
//   NULL,   
void CameraAttachToObject(GameObject* obj)
{
	attached_object = obj;
}

void SetCameraAttachedObjectOffset(float x, float y)
{
	CAMERA_OBJ_OFF_X = x;
	CAMERA_OBJ_OFF_Y = y;

	if ( attached_object )
	{
		CAMERA_X = attached_object->aabb.p.x + x;
		CAMERA_Y = attached_object->aabb.p.y + y;
	}
	else
	{
		CAMERA_X += x;
		CAMERA_Y += y;
	}
}

//  
GameObject* GetCameraAttachedObject()
{
	return attached_object;
}

//  
// x, y -  ,     
void CameraMoveToPos(float x, float y)
{
	attached_object = NULL;

	CAMERA_X = x;
	CAMERA_Y = y;

	CalcCamOffX();
	CalcCamOffY();
}

// ,      
void CameraUseBounds(bool b)
{
	camera_use_bounds = b;
}

//   
//   ,      ,      
void CameraSetBounds(float left, float right, float top, float bottom)
{
	CAMERA_LEFT_BOUND = left;
	CAMERA_RIGHT_BOUND = right;
	if (right - left < cfg.scr_width) right = left + cfg.scr_width;
	CAMERA_TOP_BOUND = top;
	CAMERA_BOTTOM_BOUND = bottom;
	if (bottom - top < cfg.scr_height) bottom = top + cfg.scr_height;
}

//   
void CameraUpdatePosition()
{
	CAMERA_OLD_OFF_X = CAMERA_OFF_X;
	CAMERA_OLD_OFF_Y = CAMERA_OFF_Y;

	if (attached_object)
	{
		if (camera_attach_x)
		{
			CAMERA_X = attached_object->aabb.p.x - CAMERA_OBJ_OFF_X;

			switch (CAMERA_FOCUS_ON_OBJ_POS)
			{
			case CamFocusLeftCenter:
				CAMERA_X -= attached_object->aabb.W;
				break;
			case CamFocusRightCenter:
				CAMERA_X += attached_object->aabb.W;
				break;
			case CamFocusLeftTopCorner:
				CAMERA_X -= attached_object->aabb.W;
				break;
			case CamFocusLeftBottomCorner:
				CAMERA_X -= attached_object->aabb.W;
				break;
			case CamFocusRightTopCorner:
				CAMERA_X += attached_object->aabb.W;
				break;
			case CamFocusRightBottomCorner:
				CAMERA_X += attached_object->aabb.W;
				break;
			default: break;
			}

			CAMERA_X = floor( CAMERA_X + 0.5f );

			CalcCamOffX();
		}

		if (camera_attach_y)
		{
			CAMERA_Y = attached_object->aabb.p.y - CAMERA_OBJ_OFF_Y;
			if (attached_object->type == objPlayer) CAMERA_Y += ((ObjPlayer*)attached_object)->camera_offset;

			switch (CAMERA_FOCUS_ON_OBJ_POS)
			{
			case CamFocusBottomCenter:
				CAMERA_Y += attached_object->aabb.H;
				break;
			case CamFocusTopCenter:
				CAMERA_Y -= attached_object->aabb.H;
				break;
			case CamFocusLeftTopCorner:
				CAMERA_Y -= attached_object->aabb.H;
				break;
			case CamFocusLeftBottomCorner:
				CAMERA_Y += attached_object->aabb.H;
				break;
			case CamFocusRightTopCorner:
				CAMERA_Y -= attached_object->aabb.H;
				break;
			case CamFocusRightBottomCorner:
				CAMERA_Y += attached_object->aabb.H;
				break;
			default: break;
			}

			CAMERA_Y = floor( CAMERA_Y + 0.5f );

			CalcCamOffY();
		}
	}

	if (camera_use_bounds)
		ApplyBounds();

	//     OpenGL -    , 
	//   .

	glTranslated( CAMERA_OFF_X+0.375, CAMERA_OFF_Y+0.375, 0);
}
