#ifndef __OBJECT_CHARACTER_H_
#define __OBJECT_CHARACTER_H_

#include "object_dynamic.h"
#include "weapon.h"

enum CharacterGunDirection { cgdNone, cgdUp, cgdDown };
enum CharacterShootingBeh { csbNoShooting, csbFreeShooting, csbOnAnimCommand };
enum CharacterOffscreenBeh { cobNone, cobDie, cobAnim, cobSleep };
extern UINT internal_time;


//     ,      
// GameObject::ProcessSprite(),    ObjCharacter* ch.
class ObjCharacter : public ObjDynamic
{
public:
	bool is_invincible;
	short health;
	short health_max;

	CharacterOffscreenBeh offscreen_behavior;
	float offscreen_distance;

	//     ,       .
	float damage_mult;

	CharacterGunDirection gunDirection;

	UINT gunReloadTick;
	UINT gunReloadTime;

	Weapon* weapon;
	Weapon* cur_weapon;

	CharacterShootingBeh shootingBeh;

	ObjCharacter()
	{
		is_invincible = false;
		health = 0;
		gunDirection = cgdNone;
		mem_anim = NULL;
		mem_frame = mem_tick = 0;
		offscreen_behavior = cobNone;
		offscreen_distance = 0.0f;

		weapon = NULL;
		cur_weapon = NULL;

		shootingBeh = csbNoShooting;

		gunReloadTick = 0;
		gunReloadTime = 0;

		damage_mult = 1.0;
	}

	virtual ~ObjCharacter()
	{
		DELETESINGLE(weapon);
		DELETEARRAY(mem_anim);
	}
	
	void ProcessShooting();

	void ReceiveDamage(UINT damage, UINT damage_type)
	{
		if ( is_invincible || activity == oatDying ) return;

		//     ,    .     .
		//TODO:    
		//             .
		if ( sprite->GetAnimation("pain") == NULL || this->activity == oatMorphing )
		{
			ReduceHealth((UINT)(damage*damage_mult));
			return;
		}

		if ( sprite->cur_anim != "pain" &&  sprite->cur_anim != "die" && damage > 0 )
		{
			DELETEARRAY(mem_anim);
			mem_anim = StrDupl(this->sprite->cur_anim.c_str());
			mem_frame = sprite->GetAnimation(this->sprite->cur_anim_num)->current;
			if ( mem_tick == 0 ) mem_tick = internal_time;
			sprite->stack->Push( (int)(damage*damage_mult) );
			sprite->stack->Push( damage_type );
			SetAnimation("pain", true );
		}

		if (health <= 0 && activity != oatDying)
		{
			SetAnimation("die", false);
			activity = oatDying;
		}
	}

	bool ReduceHealth(UINT damage)
	{
		health -= damage;
		if (health <= 0 && activity != oatDying)
		{
			SetAnimation("die", false);
			activity = oatDying;
			return true;
		}
		return false;
	}

	void ReceiveHealing(UINT amount)
	{
		health += amount;
		if ( health > health_max ) health = health_max;
	}

};

#endif // __OBJECT_CHARACTER_H_
