#include "StdAfx.h"
#include "object_ribbon.h"
#include "../camera.h"
#include "../sprite.h"
#include "../../render/texture.h"
#include "../../resource_mgr.h"
#include "../../config.h"

extern float CAMERA_BOTTOM;
extern float CAMERA_TOP;
extern float CAMERA_LEFT;
extern float CAMERA_RIGHT;
extern float CAMERA_X;
extern float CAMERA_Y;
extern ResourceMgr<Proto> * protoMgr;
extern ResourceMgr<Texture> * textureMgr;

float ObjRibbon::getBound( int num )
{
	switch (num)
	{
	case 1:
		return ubl ? max( bl + k.x*CAMERA_X, CAMERA_LEFT ) : CAMERA_LEFT;
	case 2:
		return ubr ? min( br + k.x*CAMERA_X, CAMERA_RIGHT ) : CAMERA_RIGHT;
	case 3:
		return ubt ? max( bt + k.y*CAMERA_Y, CAMERA_TOP ) : CAMERA_TOP;
	default:
		return ubb ? min( bb + k.y*CAMERA_Y, CAMERA_BOTTOM ) : CAMERA_BOTTOM;
	}
}

extern config cfg;

void ObjRibbon::Draw()
{
	//   ,   ...
	if ( this->sprite == NULL )
		return;

#ifdef MAP_EDITOR
	RenderSprite( aabb.p.x-aabb.W, aabb.p.y-aabb.H, 1, 0, 0, 2*aabb.W, 2*aabb.H, NULL, false, RGBAf(DEBUG_WAYPOINT_COLOR) );
	if ( this->show_border )
	{
		RenderLine( CAMERA_LEFT, getBound(3), CAMERA_RIGHT, getBound(3), 1, RGBAf(1, 0, 0, 1));
		RenderLine( CAMERA_LEFT, getBound(4), CAMERA_RIGHT, getBound(4), 1, RGBAf(0, 1, 0, 1));
		RenderLine( getBound(1), CAMERA_TOP, getBound(1), CAMERA_BOTTOM, 1, RGBAf(0, 0, 1, 1));
		RenderLine( getBound(2), CAMERA_TOP, getBound(2), CAMERA_BOTTOM, 1, RGBAf(1, 1, 0, 1));
	}
#endif //MAP_EDITOR

	UINT framenum = this->sprite->currentFrame;
	FrameInfo* f = (this->sprite->tex->frame + framenum);
	this->sprite->frameWidth = f->size.x;
	this->sprite->frameHeight = f->size.y;

	//,       .
	if ( ubr && getBound(1) < CAMERA_LEFT )
		return;
	if ( ubl && getBound(2) > CAMERA_RIGHT )
		return;
	if ( ubb && getBound(3) < CAMERA_TOP )
		return;
	if ( ubt && getBound(4) > CAMERA_BOTTOM )
		return;

	Vector2 coord = aabb.p + Vector2( k.x*CAMERA_X, k.y*CAMERA_Y ) - Vector2( aabb.W, aabb.H );
	coord -= Vector2( this->sprite->realX, this->sprite->realY );
	if ( !repeat_x && (coord.x + sprite->frameWidth < getBound(1) || coord.x > getBound(2)) )
		return;
	if ( !repeat_y && (coord.y + sprite->frameHeight < getBound(3) || coord.y > getBound(4)) )
		return;
	//     -    .
	if ( !repeat_x && (coord.x + this->sprite->frameWidth < CAMERA_LEFT || coord.x - this->sprite->frameWidth > CAMERA_RIGHT) )
		return;
	if ( !repeat_y && (coord.y + this->sprite->frameHeight < CAMERA_TOP || coord.y - this->sprite->frameHeight > CAMERA_BOTTOM) )
		return;

	//""     ,   .
	if ( repeat_x )
	{
		while ( coord.x + this->sprite->frameWidth < getBound(2) )
			coord.x += this->sprite->frameWidth;
		while ( coord.x + this->sprite->frameWidth > getBound(1) )
			coord.x -= this->sprite->frameWidth;
	}
	if ( repeat_y  )
	{
		while ( coord.y + this->sprite->frameHeight < getBound(4) )
			coord.y += this->sprite->frameHeight;
		while ( coord.y + this->sprite->frameHeight > getBound(3) )
			coord.y -= this->sprite->frameHeight;
	}
	Vector2 bs = Vector2( max(getBound(1) - coord.x - this->sprite->frameWidth, 0.0f), max(getBound(3) - coord.y - this->sprite->frameHeight, 0.0f) );
	if ( repeat_x ) coord.x = getBound(1);
	if ( repeat_y ) coord.y = getBound(3);
	if ( !repeat_x && coord.x < getBound(1) )
	{
		bs.x += getBound(1) - coord.x;
		coord.x = getBound(1);
	}

	Vector2 edge = Vector2( getBound(2), getBound(4) );
	float x1, x2, y1, y2;
	float rx, ry;
	bool first_x = true;
	bool first_y = true;
	for ( ry = coord.y; ry < edge.y; ry += this->sprite->frameHeight )
	{
		first_x = true;
		for ( rx = coord.x; rx < edge.x; rx += this->sprite->frameWidth )
		{
			if ( this->sprite->tex )
			{
				x1 = f->coord->x * this->sprite->tex->width;
				y1 = (1 - f->coord->y) * this->sprite->tex->height;
				if ( first_x )
				{
					x1 += bs.x;
					x2 = x1 + min( f->size.x - bs.x, edge.x - rx );
				}
				else
					x2 = x1 + min( f->size.x, edge.x - rx );
				if ( first_y )
				{
					y1 += bs.y;
					y2 = y1 + min( f->size.y - bs.y, edge.y - ry );
				}
				else
					y2 = y1 + min( f->size.y, edge.y - ry );
				
			}
			else
			{
				x1 = y1 = 0;
				x2 = 2*this->aabb.W;
				y2 = 2*this->aabb.H;
			}
#ifdef MAP_EDITOR
			if ( this->show_border )
			{
				RenderLine( aabb.p.x, aabb.p.y, rx+(x2-x1)/2, ry+(y2-y1)/2, 1, RGBAf(DEBUG_WAYPOINT_COLOR) );
			}
#endif
			RenderSprite( rx, ry, this->sprite->z, x1, y1, x2, y2, this->sprite->tex, this->sprite->IsMirrored(), this->sprite->color );
			if ( first_x )
			{
				rx -= bs.x;
				first_x = false;
			}
			if ( !repeat_x ) break;
		}
		if ( first_y )
		{
			ry -= bs.y; //    .
			first_y = false;
		}
		if ( !repeat_y ) break;
	}
}

void ObjRibbon::setBounds( float x1, float y1, float x2, float y2 )
{
	bl = x1;
	bt = y1;
	br = x2;
	bb = y2;
}

void ObjRibbon::setBoundsUsage( bool x1, bool y1, bool x2, bool y2 )
{
	ubl = x1;
	ubr = x2;
	ubt = y1;
	ubb = y2;
}

void ObjRibbon::setRepitition(bool x, bool y)
{
	repeat_x = x;
	repeat_y = y;
}

ObjRibbon* CreateRibbon( const char* texture, Vector2 k, Vector2 coord, float z, bool from_proto )
{
	ObjRibbon* ribbon = new ObjRibbon();
	ribbon->k = k;
	ribbon->aabb.p = coord;
	ribbon->aabb.W = 16;
	ribbon->aabb.H = 16;
	if ( from_proto )
	{
		Proto* proto = protoMgr->GetByName(texture, "sprites/");
		ribbon->sprite = new Sprite( proto );
		ribbon->SetAnimation( "idle", false );
	}
	else
	{
		ribbon->sprite = new Sprite();
		ribbon->sprite->tex = textureMgr->GetByName( texture );
		ribbon->sprite->frameWidth = ribbon->sprite->tex->frame->size.x;
		ribbon->sprite->frameHeight = ribbon->sprite->tex->frame->size.y;
	}
	ribbon->sprite->z = z;

#ifdef MAP_EDITOR
	ribbon->from_proto = from_proto;
	ribbon->proto_name = new char[ strlen(texture) + 1 ];
	ribbon->creation_shift = ribbon->aabb.p - coord;
	strcpy( ribbon->proto_name, texture );
#endif //MAP_EDITOR

	AddObject( ribbon );

	return ribbon;
}
