#include "StdAfx.h"

#include "player.h"
#include "camera.h"
#include "../config.h"
#include "../script/api.h"

#include "../input_mgr.h"

#include "../resource_mgr.h"

//////////////////////////////////////////////////////////////////////////


extern ResourceMgr<Proto> * protoMgr;

extern InputMgr inpmgr;

extern config cfg;

extern lua_State* lua;

extern game::GameStates game_state;

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

//    .     .
//     CreatePlayer  object_player.cpp
ObjPlayer* Player::Create(const char* proto_name, Vector2 coord, const char* start_anim)
{
	if (this->first && this->second)		//    
		return NULL;

	ObjPlayer* plr = CreatePlayer(proto_name, coord, start_anim);
	if (plr)
	{
		if (!this->first)
		{
			this->first = plr;
			this->current = plr;

			first_proto = new char[strlen(proto_name) + 1];
			memset(first_proto, 0, strlen(proto_name) + 1);
			strcpy(first_proto, proto_name);

			revivePoint = coord;		//  ,       
		}
		else
		{
			this->second = plr;
			second_proto = new char[strlen(proto_name) + 1];
			memset(second_proto, 0, strlen(proto_name) + 1);
			strcpy(second_proto, proto_name);

			//             
			plr->sprite->ClearVisible();
			plr->SetSleep();
			plr->ClearBulletCollidable();
		}
	}
	
	return plr;
}

//    num,    start_anim
ObjPlayer* Player::Revive(int num, const char* start_anim)
{
	char* pname = NULL;

	if (num == 1) 
	{
		pname = this->first_proto;
		if (this->first)
			return NULL;	//   
	}
	else if (num == 2) 
	{
		pname = this->second_proto;
		if (this->second)
			return NULL;	//   
	}
	else
		ASSERT(false);		//   

	if (!pname)
		return NULL;		// ,  

	ObjPlayer* plr = CreatePlayer(pname, revivePoint, start_anim);

	ObjPlayer* another_plr = NULL;
	const Proto* aproto = NULL;

	if (num == 1) 
	{
		this->first = plr;
		another_plr = second;
		aproto = protoMgr->GetByName(second_proto);
	}
	else if (num == 2) 
	{
		this->second = plr;
		another_plr = first;
		aproto = protoMgr->GetByName(first_proto);
	}

	//     
	if (another_plr && aproto)
	{
		if (another_plr->health < (short)aproto->health)
			another_plr->health = aproto->health;
	}

	if (!this->current) this->current = plr;
	this->current->recovery_time = 100;
	this->current->is_invincible = true;

	return plr;
}

ObjPlayer* Player::GetByNum(int num)
{
	ObjPlayer* plr = NULL;
	if (num == 1)
		plr = this->first;
	else if (num == 2)
		plr = this->second;

	return plr;
}

//    .     
void Player::Update()
{
	if (!this->first && !this->second)
		return;

	//if (!this->current)
	//	this->current = this->first;


	Changer( false, false );

	//           
}


//   
void Player::Changer( bool forced, bool immediate )
{
	if (!(this->current && this->first && this->second))
		return;		//  

	if ( !forced )
	{
		if (this->change_blocked )
			return;		//  

		if ( this->current->activity == oatDying || !this->current->IsOnPlane() )
			return;		//        ,  

	}
	if ( (forced && !immediate) || inpmgr.IsProlongedReleased(cakChangePlayer) )
	{
		this->current->vel = Vector2(0,0);
		this->current->activity = oatMorphing;
		this->current->SetAnimation("morph_out", false);
	}
	if ( (this->current->activity == oatMorphing && this->current->sprite->IsAnimDone()) || immediate )
	{
		ObjPlayer* old = this->current;
		if (this->current == this->first)
			this->current = this->second;
		else
			this->current = this->first;

		old->SetAnimation("idle", true);

		this->current->aabb.p = old->aabb.p;
		this->current->old_aabb = old->old_aabb;
		this->current->vel = old->vel;
		this->current->acc = old->acc;

		this->current->controlEnabled = old->controlEnabled;
		this->current->gravity = old->gravity;
		this->current->movement = old->movement;
		if ( old->sprite->IsMirrored() ) this->current->sprite->SetMirrored();
		else this->current->sprite->ClearMirrored();
		if (!immediate) 
		{
			this->current->SetAnimation("morph_in", false);
			this->current->activity = oatMorphing;
		}
		if (this->current->aabb.H != old->aabb.H)
			this->current->aabb.p.y = old->aabb.Bottom() - this->current->aabb.H;

		old->IsOnPlane() ? this->current->SetOnPlane() : this->current->ClearOnPlane();
		this->current->suspected_plane = old->suspected_plane;

		old->IsBulletCollidable() ?
			this->current->SetBulletCollidable() :
			this->current->ClearBulletCollidable();

		old->ClearBulletCollidable();

		this->current->movementDirectionX = old->movementDirectionX;
		/*old->sprite->IsMirrored() ?
			_PLAYER->sprite->SetMirrored() :
			_PLAYER->sprite->ClearMirrored();*/

		this->current->sprite->SetVisible();
		old->sprite->ClearVisible();

		this->current->ClearSleep();
		old->SetSleep();

		if (GetCameraAttachedObject() == old)
			CameraAttachToObject(this->current);

		if (onChangePlayerProcessor >= 0)
		{
			lua_pushinteger(lua, old->id);
			lua_pushinteger(lua, current->id);
			if (SCRIPT::ExecChunkFromReg(onChangePlayerProcessor, 2))
			{
				//    - .      .
				SCRIPT::RemoveFromRegistry(onChangePlayerProcessor);
				onChangePlayerProcessor = LUA_NOREF;
			}
		}
	}
}

//   
void Player::BlockChange()
{
	this->change_blocked = true;
}

//   
void Player::UnblockChange()
{
	this->change_blocked = false;
}

//     
void Player::OnDying(ObjPlayer* plr)
{
	if (game_state == game::GAME_DESTROYING)
		return;			//          

	if (plr == this->current)
		current = NULL;

	int num = 0;

	if (plr == this->second)
	{
		num = 2;
		this->second = NULL;
		//DELETEARRAY(this->second_proto);
	}
	else if (plr == this->first)
	{
		num = 1;
		this->first = NULL;
		//DELETEARRAY(this->first_proto);

		//if (this->second)
		//{
		//	this->first = this->second;
		//	this->first_proto = this->second_proto;
		//	this->second = NULL;
		//	this->second_proto = NULL;
		//}
	}
	else
		ASSERT(false);		//    

	if (onPlayerDeathProcessor >= 0)
	{
		PushObject(lua, plr);			//          
		lua_pushinteger(lua, num);		//   

		if (SCRIPT::ExecChunkFromReg(onPlayerDeathProcessor, 2))
		{
			//    - .      .
			SCRIPT::RemoveFromRegistry(onPlayerDeathProcessor);
			onChangePlayerProcessor = LUA_NOREF;
		}
	}
}
