#include "StdAfx.h"

#include "texture.h"
#include "renderer.h"


#include "../misc.h"

const size_t BLOCK_INIT_SIZE	= 1024;		// bytes
const size_t BLOCK_SIZE			= 512;		// bytes

#ifdef DEBUG_PRINT
extern float DBG_PRINT_VAL1;
#endif // DEBUG_PRINT

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

render_data* sprites_render_data = NULL;	//    
size_t sprites_render_data_size = 0;			//      

render_data* lines_render_data = NULL;		//    
size_t lines_render_data_size = 0;			//      

render_data* filed_shapes_render_data = NULL;		//    
size_t filed_shapes_render_data_size = 0;			//      

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//  id      ?
render_data* _id_exists(render_data* p_data, size_t data_size, GLuint id)
{
	if(data_size == 0)
		return NULL;

	assert(p_data != NULL);

	for(size_t i = 0; i < data_size; i++)
	{
		if(p_data[i].texture_id == id)
			return &p_data[i];
	}

	return NULL;

}


//   -   render_data
void _free_render_data(render_data* p_data)
{
	assert(p_data != NULL);

	if(p_data->coords)
		free(p_data->coords);

	if(p_data->vertices)
		free(p_data->vertices);

	if(p_data->colors)
		free(p_data->colors);

	p_data->vert_allocated_size = 0;
	p_data->coord_allocated_size = 0;
	p_data->colors_allocated_size = 0;
	p_data->count = 0;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

//        r_GetMem
template<typename T>
void r_ReallocArray(size_t add_size, size_t count, size_t& allocated_size, T*& arr)
{
	size_t new_size = count*sizeof(T) + add_size;
	//        
	if(new_size > allocated_size)
	{
		//      
		size_t d_size = new_size - allocated_size;

		//     BLOCK_SIZE 
		//     (  1)
		size_t blocks_to_add = (d_size / BLOCK_SIZE) + 1;

		//    
		allocated_size += (blocks_to_add * BLOCK_SIZE);

		arr = (T*)realloc(arr, allocated_size);
	}
	// TODO:    realloc  ?!

}



//         
//    ,        
//  count. 
render_data* r_GetMem(render_data** render_mas, size_t* mas_size, GLuint id, size_t vert_size, size_t coord_size, size_t color_size)
{
	//  
	render_data* x_data = _id_exists(*render_mas, *mas_size, id);

	if(x_data == NULL)
	{
		//  
		*render_mas = (render_data*)realloc(*render_mas, (*mas_size + 1)*sizeof(render_data));
		// TODO:    realloc  ?!
		//  ,     .      .
		//       .      .
		memset(&((*render_mas)[*mas_size]), 0, sizeof(render_data));
		x_data = &((*render_mas)[*mas_size]);
		x_data->texture_id = id;
		*mas_size += 1;		
	}

	{
		//   
		r_ReallocArray(vert_size, x_data->count, x_data->vert_allocated_size, x_data->vertices);
		r_ReallocArray(color_size, x_data->count, x_data->colors_allocated_size, x_data->colors);
		if (id)
			r_ReallocArray(coord_size, x_data->count, x_data->coord_allocated_size, x_data->coords);
	}

	return x_data;
}


//  ,   
void r_ZeroRenderData(void)
{
	for(size_t i = 0; i < sprites_render_data_size; i++)
	{
		sprites_render_data[i].count = 0;
	}

	for(size_t i = 0; i < lines_render_data_size; i++)
	{
		lines_render_data[i].count = 0;
	}

	for(size_t i = 0; i < filed_shapes_render_data_size; i++)
	{
		filed_shapes_render_data[i].count = 0;
	}


#ifdef _DEBUG_RENDER
	LogToFile("Render data was zeroized");
#endif // _DEBUG_RENDER
}

//   ( )
void r_ClearRenderData(void)
{
	for(size_t i = 0; i < sprites_render_data_size; i++)
	{
		_free_render_data(&sprites_render_data[i]);
	}
	free(sprites_render_data);
	sprites_render_data = NULL;
	sprites_render_data_size = 0;

	for(size_t i = 0; i < lines_render_data_size; i++)
	{
		_free_render_data(&lines_render_data[i]);
	}
	free(lines_render_data);
	lines_render_data = NULL;
	lines_render_data_size = 0;

	for(size_t i = 0; i < filed_shapes_render_data_size; i++)
	{
		_free_render_data(&filed_shapes_render_data[i]);
	}
	free(filed_shapes_render_data);
	filed_shapes_render_data = NULL;
	filed_shapes_render_data_size = 0;
}

void RenderSprites()
{
	if(!sprites_render_data || !sprites_render_data_size)
		return;

	glEnable(GL_TEXTURE_2D);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	render_data* null_tex = NULL;

	for(size_t i = 0; i < sprites_render_data_size; i++)
	{
		render_data* rdata = &sprites_render_data[i];

		if (rdata->texture_id)
		{
			glBindTexture(GL_TEXTURE_2D, rdata->texture_id);
			glTexCoordPointer(2, GL_FLOAT, 0, rdata->coords);
			glVertexPointer(3, GL_FLOAT, 0, rdata->vertices);
			glColorPointer(4, GL_FLOAT, 0, rdata->colors);
			glDrawArrays(GL_QUADS, 0, rdata->count);
		}
		else
		{
			null_tex = rdata;
		}
	}

	// TODO:  ,          
	if (null_tex)
	{
		//   ,      
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisable(GL_TEXTURE_2D);

		glVertexPointer(3, GL_FLOAT, 0, null_tex->vertices);
		glColorPointer(4, GL_FLOAT, 0, null_tex->colors);
		glDrawArrays(GL_QUADS, 0, null_tex->count);
	}


	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

}

void RenderLines()
{
	if(!lines_render_data || !lines_render_data_size)
		return;

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	glDisable(GL_TEXTURE_2D);

	for(size_t i = 0; i < lines_render_data_size; i++)
	{
		render_data* rdata = &lines_render_data[i];

		//glPushAttrib(GL_CURRENT_BIT);
		//glColor3f(1.0f, 0.5f, 0.25f);


		glVertexPointer(3, GL_FLOAT, 0, rdata->vertices);
		glColorPointer(4, GL_FLOAT, 0, rdata->colors);
		glDrawArrays(GL_LINES, 0, rdata->count);

		//glPopAttrib();
	}

	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}

void RenderFilledShapes()
{
	if(!filed_shapes_render_data || !filed_shapes_render_data_size)
		return;

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);

	glDisable(GL_TEXTURE_2D);

	for(size_t i = 0; i < filed_shapes_render_data_size; i++)
	{
		render_data* rdata = &filed_shapes_render_data[i];

		glVertexPointer(3, GL_FLOAT, 0, rdata->vertices);
		glColorPointer(4, GL_FLOAT, 0, rdata->colors);
		glDrawArrays(GL_TRIANGLE_FAN, 0, rdata->count);
	}

	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}

//  
void r_RenderAll(void)
{
	RenderSprites();
	RenderLines();
	RenderFilledShapes();
}

//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////

void RenderSprite(float x, float y, float z, float x1, float y1, float x2, float y2, const Texture* tex, BOOL mirrorX, const RGBAf& color)
{
#ifdef DEBUG_PRINT
	DBG_PRINT_VAL1++;
#endif // DEBUG_PRINT

	//  -  4 .
	size_t count = 4;

	render_data* rd = r_GetMem(&sprites_render_data, &sprites_render_data_size, tex ? tex->tex : 0,
		count*sizeof(vertex3f_t), tex ? count*sizeof(coord2f_t) : 0, count*sizeof(RGBAf));
	ASSERT(rd && rd->colors && rd->vertices && rd->vert_allocated_size && rd->colors_allocated_size);

	count = rd->count;
	rd->count += 4;

	rd->vertices[count].z = rd->vertices[count + 1].z = rd->vertices[count + 2].z = rd->vertices[count + 3].z = z;

	float dx = x2-x1;
	float dy = y2-y1;

	rd->vertices[count+0].x = x ;
	rd->vertices[count+0].y = y;

	rd->vertices[count+1].x = x + dx;
	rd->vertices[count+1].y = y;

	rd->vertices[count+2].x = x + dx;
	rd->vertices[count+2].y = y + dy;

	rd->vertices[count+3].x = x;
	rd->vertices[count+3].y = y + dy;

	rd->colors[count+0] = rd->colors[count+1] = rd->colors[count+2] = rd->colors[count+3] = color;

	if(tex)
	{
		ASSERT(rd->coord_allocated_size && rd->coords && rd->texture_id == tex->tex);

		float CWidth = (float)tex->width;
		float CHeight = (float)tex->height;
		float cx1 = x1 / CWidth;
		float cx2 = x2 / CWidth;
		float cy1 = 1 - y1 / CHeight;
		float cy2 = 1 - y2 / CHeight;

		if(mirrorX)
		{
			rd->coords[count+0].x = cx2;
			rd->coords[count+0].y = cy1;

			rd->coords[count+1].x = cx1;
			rd->coords[count+1].y = cy1;

			rd->coords[count+2].x = cx1;
			rd->coords[count+2].y = cy2;

			rd->coords[count+3].x = cx2;
			rd->coords[count+3].y = cy2;
		}
		else
		{
			rd->coords[count+0].x = cx1;
			rd->coords[count+0].y = cy1;

			rd->coords[count+1].x = cx2;
			rd->coords[count+1].y = cy1;

			rd->coords[count+2].x = cx2;
			rd->coords[count+2].y = cy2;

			rd->coords[count+3].x = cx1;
			rd->coords[count+3].y = cy2;
		}
	}
	else
	{
		// ,    .
		ASSERT(!rd->coords && !rd->coord_allocated_size);
	}

#ifdef DEBUG_DRAW_SPRITES_BORDERS
	RenderBox(x, y, z, x2-x1, y2-y1, RGBAf(DEBUG_SPRITES_BORDERS_COLOR));
#endif // DEBUG_DRAW_SPRITES_BORDERS

}

void RenderSprite(float x, float y, float z, coord2f_t* frame_size, coord2f_t* tex_coords, const Texture* tex, const RGBAf& color)
{
#ifdef DEBUG_PRINT
	DBG_PRINT_VAL1++;
#endif // DEBUG_PRINT

	//  -  4 .
	size_t count = 4;

	render_data* rd = r_GetMem(&sprites_render_data, &sprites_render_data_size, tex ? tex->tex : 0,
		count*sizeof(vertex3f_t), tex ? count*sizeof(coord2f_t) : 0, count*sizeof(RGBAf));
	ASSERT(rd && rd->colors && rd->vertices && rd->vert_allocated_size && rd->colors_allocated_size);

	count = rd->count;
	rd->count += 4;

	rd->vertices[count].z = rd->vertices[count + 1].z = rd->vertices[count + 2].z = rd->vertices[count + 3].z = z;

	rd->vertices[count].x = x;
	rd->vertices[count].y = y;

	rd->vertices[count+1].x = x + frame_size->x;
	rd->vertices[count+1].y = y;

	rd->vertices[count+2].x = x + frame_size->x;
	rd->vertices[count+2].y = y + frame_size->y;

	rd->vertices[count+3].x = x;
	rd->vertices[count+3].y = y + frame_size->y;

	rd->colors[count] = rd->colors[count+1] = rd->colors[count + 2] = rd->colors[count + 3] = color;

	if(tex)
	{
		ASSERT(rd->coord_allocated_size && rd->coords && rd->texture_id == tex->tex);

		memcpy(rd->coords+count, tex_coords, 4 * sizeof(coord2f_t));
	}
	else
	{
		// ,    .
		ASSERT(!rd->coords && !rd->coord_allocated_size);
	}


}


void RenderSpritePart(float x, float y, float z, coord2f_t* full_frame_size, coord2f_t* part_frame_size, coord2f_t* tex_coords, bool mirrorX, const Texture* tex, const RGBAf& color)
{
#ifdef DEBUG_PRINT
	DBG_PRINT_VAL1++;
#endif // DEBUG_PRINT
	//  -  4 .
	size_t count = 4;

	render_data* rd = r_GetMem(&sprites_render_data, &sprites_render_data_size, tex ? tex->tex : 0,
		count*sizeof(vertex3f_t), tex ? count*sizeof(coord2f_t) : 0, count*sizeof(RGBAf));
	ASSERT(rd && rd->colors && rd->vertices && rd->vert_allocated_size && rd->colors_allocated_size);

	count = rd->count;
	rd->count += 4;


	for (size_t i = count; i < count+4; i++)
	{
		rd->vertices[i].z = z;
		memcpy(rd->colors + i, &color, sizeof(RGBAf));
	}

	rd->vertices[count+0].x = x;
	rd->vertices[count+0].y = y;
	rd->vertices[count+1].x = x + part_frame_size->x;
	rd->vertices[count+1].y = y;
	rd->vertices[count+2].x = x + part_frame_size->x;
	rd->vertices[count+2].y = y + part_frame_size->y;
	rd->vertices[count+3].y = y + part_frame_size->y;
	rd->vertices[count+3].x = x;

	if(tex)
	{
		ASSERT(rd->coord_allocated_size && rd->coords && rd->texture_id == tex->tex);

		float cx1;
		float cx2;
		float cy1;
		float cy2;
		if (mirrorX)
		{
			cx1 = tex_coords[1].x; 
			cx2 = tex_coords[0].x;
		}
		else
		{
			cx1 = tex_coords[0].x; 
			cx2 = tex_coords[1].x;
		}
		cy1 = tex_coords[0].y;
		cy2 = tex_coords[3].y;

		cx2 = cx1 + part_frame_size->x * (cx2 - cx1) / full_frame_size->x;
		cy2 = cy1 + part_frame_size->y * (cy2 - cy1) / full_frame_size->y;

		if(mirrorX)
		{
			rd->coords[count+0].x = cx2;
			rd->coords[count+0].y = cy1;

			rd->coords[count+1].x = cx1;
			rd->coords[count+1].y = cy1;

			rd->coords[count+2].x = cx1;
			rd->coords[count+2].y = cy2;

			rd->coords[count+3].x = cx2;
			rd->coords[count+3].y = cy2;
		}
		else
		{
			rd->coords[count+0].x = cx1;
			rd->coords[count+0].y = cy1;

			rd->coords[count+1].x = cx2;
			rd->coords[count+1].y = cy1;

			rd->coords[count+2].x = cx2;
			rd->coords[count+2].y = cy2;

			rd->coords[count+3].x = cx1;
			rd->coords[count+3].y = cy2;
		}
	}
	else
	{
		// ,    .
		ASSERT(!rd->coords && !rd->coord_allocated_size);
	}




#ifdef DEBUG_DRAW_SPRITES_BORDERS
	RenderBox(x, y, z, part_frame_size->x, part_frame_size->y, RGBAf(DEBUG_SPRITES_BORDERS_COLOR));
	if (mirrorX)
	{
		RenderBox(x - (full_frame_size->x - part_frame_size->x), y - (full_frame_size->y - part_frame_size->y), z, part_frame_size->x, part_frame_size->y, RGBAf(DEBUG_SPRITES_BORDERS_COLOR));
	}
	else
	{
		RenderBox(x, y, z, full_frame_size->x, full_frame_size->y, RGBAf(DEBUG_SPRITES_BORDERS_COLOR));
	}
#endif // DEBUG_DRAW_SPRITES_BORDERS
}


void RenderBox(float x, float y, float z, float w, float h, const RGBAf& color)
{
#ifdef DEBUG_PRINT
	DBG_PRINT_VAL1++;
#endif // DEBUG_PRINT

	//  -  8 .
	size_t count = 8;

	render_data* rd = r_GetMem(&lines_render_data, &lines_render_data_size, 0,
		count*sizeof(vertex3f_t), 0, count*sizeof(RGBAf));
	ASSERT(rd && rd->colors && rd->vertices && rd->vert_allocated_size && rd->colors_allocated_size);

	count = rd->count;
	rd->count += 8;

	ASSERT(!rd->coords && !rd->coord_allocated_size && !rd->texture_id);


	rd->vertices[count+0].z = rd->vertices[count+1].z = rd->vertices[count+2].z = rd->vertices[count+3].z =
		rd->vertices[count+4].z = rd->vertices[count+5].z = rd->vertices[count+6].z = rd->vertices[count+7].z = z;

	rd->colors[count+0] = rd->colors[count+1] = rd->colors[count+2] = rd->colors[count+3] =
		rd->colors[count+4] = rd->colors[count+5] = rd->colors[count+6] = rd->colors[count+7] = color;

	// 1 - - 2
	//glVertex2i(_x, y);
	rd->vertices[count+0].x = x;
	rd->vertices[count+0].y = y;
	//glVertex2i(x + w, y);
	rd->vertices[count+1].x = x + w;
	rd->vertices[count+1].y = y;

	// 2 - - 3
	//glVertex2i(x + w, y);
	rd->vertices[count+2].x = x + w;
	rd->vertices[count+2].y = y;
	//glVertex2i(x + w, y + h);
	rd->vertices[count+3].x = x + w;
	rd->vertices[count+3].y = y + h;

	// 3 - - 4
	//glVertex2i(x + w, y + h);
	rd->vertices[count+4].x = x + w;
	rd->vertices[count+4].y = y + h;
	//glVertex2i(x, y + h);
	rd->vertices[count+5].x = x;
	rd->vertices[count+5].y = y + h;

	// 4 - - 1
	//glVertex2i(_x, y + h);
	rd->vertices[count+6].x = x;
	rd->vertices[count+6].y = y + h;
	//glVertex2i(x, y);
	rd->vertices[count+7].x = x;
	rd->vertices[count+7].y = y;

}

void RenderLine(float x1, float y1, float x2, float y2, float z, const RGBAf& color)
{
#ifdef DEBUG_PRINT
	DBG_PRINT_VAL1++;
#endif // DEBUG_PRINT

	//  - 2 .
	size_t count = 2;

	render_data* rd = r_GetMem(&lines_render_data, &lines_render_data_size, 0,
		count*sizeof(vertex3f_t), 0, count*sizeof(RGBAf));
	ASSERT(rd && rd->colors && rd->vertices && rd->vert_allocated_size && rd->colors_allocated_size);

	count = rd->count;
	rd->count += 2;

	ASSERT(!rd->coords && !rd->coord_allocated_size && !rd->texture_id);

	rd->vertices[count+0].z = rd->vertices[count+1].z = z;

	rd->colors[count+0] = rd->colors[count+1] = color;

	// 1 - - 2
	//glVertex2i(_x, y);
	rd->vertices[count+0].x = x1;
	rd->vertices[count+0].y = y1;
	//glVertex2i(x + w, y);
	rd->vertices[count+1].x = x2;
	rd->vertices[count+1].y = y2;

}


void RenderEllipse(float x, float y, float w, float h, float z, const RGBAf& color, UINT id, UINT segments)
{
#ifdef DEBUG_PRINT
	DBG_PRINT_VAL1++;
#endif // DEBUG_PRINT

	render_data* rd = r_GetMem(&filed_shapes_render_data, &filed_shapes_render_data_size, id,
		segments*sizeof(vertex3f_t), 0, segments*sizeof(RGBAf));
	ASSERT(rd && rd->colors && rd->vertices && rd->vert_allocated_size && rd->colors_allocated_size);

	//  id   ,        .
	rd->texture_id = id;


	rd->coords = NULL;
	rd->coord_allocated_size = 0;

	size_t count = rd->count;
	for (GLfloat i = 0; i < 2*M_PI && count < rd->count + segments; i+=(2*(GLfloat)M_PI / segments), count++)
	{
		rd->vertices[count].x = x + cos(i) * w;
		rd->vertices[count].y = y + sin(i) * h;
		rd->vertices[count].z = z;
		rd->colors[count] = color;
	}
	rd->count = count;

}

