#include "StdAfx.h"

#include "../gui/gui.h"



#include "api.h"

//////////////////////////////////////////////////////////////////////////

extern Gui* gui;
extern lua_State* lua;

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


int scriptApi::CreateWidget(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, "  ");
	luaL_argcheck(L, lua_isstring(L, 2), 2, "  ");
	luaL_argcheck(L, lua_islightuserdata(L, 3) || lua_isnil(L, 3), 3, " ");
	luaL_argcheck(L, lua_isnumber(L, 4), 4, " x");
	luaL_argcheck(L, lua_isnumber(L, 5), 5, " y");
	luaL_argcheck(L, lua_isnumber(L, 6), 6, " w");
	luaL_argcheck(L, lua_isnumber(L, 7), 7, " h");

	WidgetTypes wt = (WidgetTypes)lua_tointeger(L, 1);
	const char* name = lua_tostring(L, 2);
	//void* parent = lua_touserdata(L, 3);
	float x = (float)lua_tonumber(L, 4);
	float y = (float)lua_tonumber(L, 5);
	float w = (float)lua_tonumber(L, 6);
	float h = (float)lua_tonumber(L, 7);

	UINT id = gui->CreateWidget(wt, name, x, y, w, h);
	lua_pop(L, lua_gettop(L));
	lua_pushinteger(L, id);
	return 1;
}

int scriptApi::DestroyWidget(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	gui->DestroyWidget((UINT)lua_tointeger(L, 1));
	return 0;
}


int scriptApi::WidgetSetBorder(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isboolean(L, 2), 2, " bool");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
		wi->border = lua_toboolean(L, 2) != 0;
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetBorder:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}


int scriptApi::WidgetSetCaption(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isstring(L, 2), 2, " caption");
	luaL_argcheck(L, lua_isboolean(L, 3)||lua_isnil(L, 3)||lua_isnone(L, 3), 2, " bool multiline");

	bool multiline = false;
	if ( lua_isboolean(L, 3) ) multiline = lua_toboolean(L, 3) != 0;

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
		wi->SetCaption(lua_tostring(L, 2), multiline);
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetCaption:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}
	return 0;
}

int scriptApi::WidgetGetCaption(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
		lua_pushstring(L, wi->caption);
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetGetCaption:  id=%d  ", (UINT)lua_tointeger(L, 1));
		lua_pushnil(L);
	}
	return 1;
}

int scriptApi::WidgetGetName(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
		lua_pushstring(L, wi->name);
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetGetName:  id=%d  ", (UINT)lua_tointeger(L, 1));
		lua_pushnil(L);
	}
	return 1;
}

int scriptApi::WidgetSetCaptionColor(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_istable(L, 2), 2, "  color");
	luaL_argcheck(L, lua_isboolean(L, 3), 3, " bool");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		if (lua_toboolean(L, 3))
			SCRIPT::GetColorFromTable(L, 2, wi->caption_act_color);
		else
			SCRIPT::GetColorFromTable(L, 2, wi->caption_inact_color);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetCaptionColor:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}
	return 0;
}

int scriptApi::WidgetSetBorderColor(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_istable(L, 2), 2, "  color");
	luaL_argcheck(L, lua_isboolean(L, 3), 3, " bool");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		if (lua_toboolean(L, 3))
			SCRIPT::GetColorFromTable(L, 2, wi->border_active_color);
		else
			SCRIPT::GetColorFromTable(L, 2, wi->border_inactive_color);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetBorderColor:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetCaptionFont(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isstring(L, 2), 2, " fontname");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		wi->caption_font = FontByName(lua_tostring(L, 2));
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetCaptionFont:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::GetCaptionSize(lua_State* L)
{
	luaL_argcheck(L, lua_isstring(L, 1), 1, " ");
	luaL_argcheck(L, lua_isstring(L, 2), 2, " ");

	Font* font = FontByName(lua_tostring(L, 1));
	int result_y = font->GetStringHeight(lua_tostring(L, 2));
	int result_x = font->GetStringWidth(lua_tostring(L, 2));
	lua_pushinteger(L, result_x);
	lua_pushinteger(L, result_y);
	return 2;
}

int scriptApi::WidgetSetVisible(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isboolean(L, 2), 2, " bool");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		wi->visible = lua_toboolean(L, 2) != 0;
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetVisible:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetGetVisible(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		lua_pushboolean(L, wi->visible);
	}
	else
	{
		lua_pushnil(L);
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetGetVisible:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 1;
}

int scriptApi::WidgetSetFixedPosition(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isboolean(L, 2), 2, " bool");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		wi->SetFixedPosition(lua_toboolean(L, 2) != 0);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetFixedPosition:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetGetSize(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		lua_pushnumber(L, 2*wi->aabb.W);
		lua_pushnumber(L, 2*wi->aabb.H);
	}
	else
	{
		lua_pushnil(L);
		lua_pushnil(L);
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetGetSize:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 2;
}

int scriptApi::WidgetGetPos(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		lua_pushnumber(L, wi->aabb.Left());
		lua_pushnumber(L, wi->aabb.Top());
	}
	else
	{
		lua_pushnil(L);
		lua_pushnil(L);
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetGetSize:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 2;
}

int scriptApi::WidgetSetFocusable(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isboolean(L, 2), 2, " bool");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		wi->focusable = lua_toboolean(L, 2) != 0;
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetFocusable:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}



int scriptApi::WidgetSetLMouseClickProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onMouseLClickProc, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetLMouseClickProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetRMouseClickProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onMouseRClickProc, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetRMouseClickProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetMouseEnterProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	
	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onMouseEnterProc, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetMouseEnterProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetMouseLeaveProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onMouseLeaveProc, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetMouseLeaveProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetKeyDownProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onKeyDown, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetKeyDownProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetKeyPressProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onKeyPress, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetKeyPressProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetKeyInputProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onKeyInput, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetKeyInputProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetFocusProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onFocus, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetFocusProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetUnFocusProc(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		SCRIPT::RegProc(L, &wi->onUnFocus, 2);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetUnFocusProc:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetGainFocus(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	UINT id = (UINT)lua_tointeger(L, 1);

	gui->SetFocus(id);

	return 0;
}

int scriptApi::WidgetBringToFront(lua_State* L)
{
	luaL_error(L, "WidgetBringToFront  ");
	return 0;
	//luaL_argcheck(L, lua_islightuserdata(L, 1), 1, " ");

	//void* p = lua_touserdata(L, 1);
	//GuiWidget* wi = (GuiWidget*)p;

	//wi->BringToFront();

	//return 0;
}

int scriptApi::WidgetSetSize(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isnumber(L, 2), 2, " ");
	luaL_argcheck(L, lua_isnumber(L, 3), 3, " ");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		float w = (float)lua_tonumber(L, 2);
		float h = (float)lua_tonumber(L, 3);

		float top = wi->aabb.Top();
		float left = wi->aabb.Left();
		wi->aabb = CAABB(left, top, left + w, top + h);
		// TODO:     ,  .  .
		//wi->aabb.W = w/2;
		//wi->aabb.H = w/2;
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetSize:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}


	return 0;
}

int scriptApi::WidgetSetPos(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isnumber(L, 2), 2, " x1");
	luaL_argcheck(L, lua_isnumber(L, 3), 3, " y1");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		float x1 = (float)lua_tonumber(L, 2);
		float y1 = (float)lua_tonumber(L, 3);

		float top = wi->aabb.Top();
		float left = wi->aabb.Left();
		float bottom = wi->aabb.Bottom();
		float right = wi->aabb.Right();
		wi->aabb = CAABB(x1, y1, x1 + (right - left), y1 + (bottom - top));
		// TODO:     ,  .  .
		//wi->aabb.p.x = x1 - wi->aabb.W;
		//wi->aabb.p.y = y1 - wi->aabb.H;
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetPos:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetZ(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isnumber(L, 2), 2, " z");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		wi->SetZ((float)lua_tonumber(L, 2) );
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetZ:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetSprite(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isstring(L, 2), 2, " proto_name");
	luaL_argcheck(L, lua_isstring(L, 3) || lua_isnone(L, 3), 3, " start_anim");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi)
	{
		wi->SetSprite(lua_tostring(L, 2), lua_tostring(L, 3));
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetSprite:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}

int scriptApi::WidgetSetColorBox(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_istable(L, 2), 2, "  color");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	RGBAf color;
	SCRIPT::GetColorFromTable(L, 2, color);
	if (wi)
	{
		wi->SetColorBox(color);
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetColorBox:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}
	return 0;
}

int scriptApi::WidgetSetAnim(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	luaL_argcheck(L, lua_isstring(L, 2), 2, " anim_name");

	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (wi && wi->sprite)
	{
		wi->sprite->SetAnimation(string(lua_tostring(L, 2)));
	}
	else
	{
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetAnim:  id=%d  ", (UINT)lua_tointeger(L, 1));
	}

	return 0;
}


int scriptApi::GlobalSetKeyDownProc(lua_State* L)
{
	SCRIPT::RegProc(L, &gui->onKeyDownGlobal, 1);	
	return 0;
}

int scriptApi::GlobalSetKeyReleaseProc(lua_State* L)
{
	SCRIPT::RegProc(L, &gui->onKeyReleaseGlobal, 1);	
	return 0;
}

int scriptApi::GlobalGetKeyDownProc(lua_State* L)
{
	SCRIPT::GetFromRegistry(L, gui->onKeyDownGlobal);
	return 1;
}

int scriptApi::GlobalGetKeyReleaseProc(lua_State* L)
{
	SCRIPT::GetFromRegistry(L, gui->onKeyReleaseGlobal);
	return 1;
}

int scriptApi::GlobalSetMouseKeyDownProc(lua_State* L)
{
	SCRIPT::RegProc(L, &gui->onMouseKeyDownGlobal, 1);	
	return 0;
}

int scriptApi::GlobalSetMouseKeyReleaseProc(lua_State* L)
{
	SCRIPT::RegProc(L, &gui->onMouseKeyReleaseGlobal, 1);	
	return 0;
}

int scriptApi::GlobalGetMouseKeyDownProc(lua_State* L)
{
	SCRIPT::GetFromRegistry(L, gui->onMouseKeyDownGlobal);
	return 1;
}

int scriptApi::GlobalGetMouseKeyReleaseProc(lua_State* L)
{
	SCRIPT::GetFromRegistry(L, gui->onMouseKeyReleaseGlobal);
	return 1;
}


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

int scriptApi::PushWidget(lua_State* L)
{
	luaL_argcheck(L, lua_isnumber(L, 1), 1, " id ");
	GuiWidget* wi = gui->GetWidget((UINT)lua_tointeger(L, 1));
	if (!wi)
	{
		lua_pushnil(L);
		sLog(DEFAULT_SCRIPT_LOG_NAME, LOG_WARNING_EV, " WidgetSetAnim:  id=%d  ", (UINT)lua_tointeger(L, 1));
		return 1;
	}

	lua_newtable(L);

	lua_pushnumber(L, wi->id);					lua_setfield(L, -2, "id");
	lua_pushstring(L, wi->name);				lua_setfield(L, -2, "name");

	lua_pushboolean(L, wi->active);				lua_setfield(L, -2, "active");
	lua_pushboolean(L, wi->visible);			lua_setfield(L, -2, "visible");
	lua_pushboolean(L, wi->dead);				lua_setfield(L, -2, "dead");

	SCRIPT::PushAABB(L, wi->aabb);				lua_setfield(L, -2, "aabb");

	if(wi->sprite)
	{
		lua_newtable(L);	// : obj sprite
		lua_pushboolean(L, wi->sprite->IsMirrored());	lua_setfield(L, -2, "mirrored");
		lua_pushboolean(L, wi->sprite->IsFixed());		lua_setfield(L, -2, "fixed");
		lua_pushboolean(L, wi->sprite->IsVisible());	lua_setfield(L, -2, "visible");
		lua_pushboolean(L, wi->sprite->IsAnimDone());	lua_setfield(L, -2, "animDone");
		lua_pushstring(L, wi->sprite->cur_anim.c_str()); lua_setfield(L, -2, "cur_anim");
		lua_setfield(L, -2, "sprite");	// : obj
	}

	lua_pushboolean(L, wi->border);					lua_setfield(L, -2, "border");


	lua_pushstring(L, wi->caption);					lua_setfield(L, -2, "caption");
	if (wi->caption_font)
	{
		//lua_pushstring(L, wi->caption_font->name
	}

	lua_pushnumber(L, wi->lastMousePos);			lua_setfield(L, -2, "lastMousePos");

	{
		lua_newtable(L);
		lua_pushnumber(L, wi->lastmouseButtonState[0]);			lua_setfield(L, -2, "left");
		lua_pushnumber(L, wi->lastmouseButtonState[1]);			lua_setfield(L, -2, "right");
		lua_pushnumber(L, wi->lastmouseButtonState[2]);			lua_setfield(L, -2, "middle");
		lua_setfield(L, -2, "lastmouseButtonState");
	}


	return 1;
}

