#include "StdAfx.h"

#include "script.h"

#include "luathread.h"
#include "timerevent.h"
#include "../game/game.h"

//////////////////////////////////////////////////////////////////////////

extern UINT internal_time;
extern UINT last_timerevent_tick;

//////////////////////////////////////////////////////////////////////////
//  
list <TimerEvent*> timerEvents;
UINT nbPausableEvents = 0;

typedef list<TimerEvent*>::iterator TEvIter;

//      ,    
// ,    .  ,   
//  ,    ,      .


//     createdEvents   
void AddTimerEvent(TimerEvent* te)
{
	if (te->IsPausable())
		nbPausableEvents++;

	//  ,   ,  .
	for(TEvIter it = timerEvents.begin();
		it != timerEvents.end();
		it++)
	{
		if (te->execTime < (*it)->execTime)
		{
			//    
			timerEvents.insert(it, te);
			return;
		}
	}
	//    ,   
	timerEvents.push_back(te);
}

//   
// dt - ,    
// action -    
// period -     
// maxCalls -  
UINT AddTimerEvent(UINT dt, int action, UINT period, UINT maxCalls)
{
	TimerEvent* te = new TimerEvent();

	te->action = action;
	te->execTime = internal_time + dt;

	te->SetPeriodic();
	te->period = period;
	te->maxCalls = maxCalls;

	AddTimerEvent(te);
	return 1;
}

//  
// dt - ,    
// action -    
UINT AddTimerEvent(UINT dt, int action)
{
	TimerEvent* te = new TimerEvent();

	te->action = action;
	te->execTime = internal_time + dt;

	AddTimerEvent(te);

	return 1;
}

//     
void AddThreadTimerEvent(UINT dt, int action, bool pausable)
{
	TimerEvent* te = new TimerEvent();
	te->action = action;
	te->execTime = internal_time + dt;
	te->SetThreadResume();
	if (pausable) te->SetPausable();

	AddTimerEvent(te);
}

//      action
int DeleteTimerEvent(int action)
{
	//       
	TimerEvent* t = NULL;
	TEvIter it;
	for(it = timerEvents.begin();
		it != timerEvents.end();
		it++)
	{
		t = *it;
		if (t->action == action)
			break;
	}

	if (it != timerEvents.end())
	{
		timerEvents.erase(it);
		if (t->IsPausable())
		{
			ASSERT(nbPausableEvents);
			nbPausableEvents--;
		}
		DELETESINGLE(t);
		return 1;
	}
	return 0;
	// TODO:     ,     
	//  .        .
	//iterator _Last = end();
	//for (iterator _First = begin(); _First != _Last; )
	//	if ((*_First)->)
	//		_First = erase(_First);
	//	else
	//		++_First;
}

//     
void DeleteAllThreadEvents()
{
	TimerEvent* t = NULL;
	TEvIter it;
	for(it = timerEvents.begin(); it != timerEvents.end(); )
	{
		t = *it;
		if (t->IsThreadResume())
		{
			DELETESINGLE(t);
			it = timerEvents.erase(it);
		}
		else
		{
			++it;
		}
	}
}

//     
void DeleteAllEvents()
{
	for(TEvIter it = timerEvents.begin();
		it != timerEvents.end();
		it++)
	{
		DELETESINGLE(*it);
	}

	timerEvents.clear();
}


//   
//     TIMEREVENTTICK 
void ProcessTimerEvents()
{
	if (timerEvents.empty())
		return;

	TimerEvent* first = NULL;

	if (nbPausableEvents && game::GetPause())
	{
		//  execTime     shift
		UINT shift = internal_time - last_timerevent_tick;


		list<TimerEvent*>::reverse_iterator rit = timerEvents.rbegin();
		list<TimerEvent*>::reverse_iterator prev_rit = rit;
		list<TimerEvent*>::reverse_iterator temp_rit = rit;
		//     
		for(; rit != timerEvents.rend(); prev_rit = rit, rit++)
		{
			if (!(*rit)->IsPausable())
				continue;
			
			first = *rit;
			first->execTime += shift;
	
			temp_rit = rit;
			//   ,    
			while((*prev_rit)->execTime < first->execTime)
			{
				*temp_rit = *prev_rit;
				*prev_rit = first;
				if (prev_rit == timerEvents.rbegin())
					break;
				temp_rit = prev_rit;
				prev_rit--;				
			}
		}
	}

	while (!timerEvents.empty())
	{
		first = *(timerEvents.begin());	//    

		if (first->execTime > internal_time)
			break;							//    

		ASSERT((!first->IsPausable() && game::GetPause()) || !game::GetPause());

		first->ExecEvent();

		if (timerEvents.empty() || 
			first != *(timerEvents.begin()))
		{
			//     -   
			continue;
		}

		if (first->IsPeriodic())
		{
			first->calls++;
			if (first->calls >= first->maxCalls - 1)
			{
				//    
				//          
				first->ClearPeriodic();
			}

			first->execTime = internal_time + first->period;
			//         
			timerEvents.pop_front();
			AddTimerEvent(first);

		}
		else
		{
			//   
			timerEvents.pop_front();
			DELETESINGLE(first);
		}
	}

}


//   .
void TimerEvent::ExecEvent()
{
	if (this->IsThreadResume())
	{
		ProcessThread(this->action);
	}
	else
	{
		SCRIPT::ExecChunkFromReg(this->action);
	}
}
